# -*- coding: utf-8 -*-
'''preggy core classes: `Assertions` and `Expect`.
'''
# preggy assertions
# https://github.com/heynemann/preggy

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2013 Bernardo Heynemann heynemann@gmail.com

import re


class Assertions(object):
    registered_assertions = {}

    @classmethod
    def assertion(cls, method):
        '''Function decorator.  Provides lower-level control for custom
        assertions than `@Assertions.create_assertions`.

        If you need more control over your error message, or your assertion
        doesn’t have a corresponding `not_`, use this decorator and
        raise a `AssertionError`.

        By raising a `AssertionError`, you get the benefit of highlighting
        the important values when your tests are broken.

        If you still just wanna raise an `AssertionError` like old times,
        that’s supported, too.

        It’s recommended to always declare both the assertion and the `not_`
        assertion (if applicable), so they can be used like this:

            expect(5).to_be_a_positive_integer()
            expect(-3).Not.to_be_a_positive_integer()
        '''
        def method_name(*args, **kw):
            method(*args, **kw)

        def exec_assertion(*args, **kw):
            return method_name(*args, **kw)

        Assertions.registered_assertions[method.__name__] = exec_assertion

        return method_name

    @classmethod
    def create_assertions(cls, method):
        '''Function decorator.  Use to create custom assertions for your
        tests.
        ''' '''
        Creating new assertions for use with `expect` is as simple as using
        this decorator on a function. The function expects `topic` as the
        first parameter, and `expectation` second:

            @Assertions.create_assertions
            def to_be_greater_than(topic, expected):
                return topic > expected

        Now, the following expectation…

            expect(2).to_be_greater_than(3)

        …will report:

            Expected topic(2) to be greater than 3.

        It will also create the corresponding `not_` assertion:

            expect(4).not_to_be_greater_than(3);

        …will report:

            Expected topic(4) not to be greater than 3.
        '''
        humanized_method_name = re.sub(r'_+', ' ', method.__name__)

        def _assertion_msg(assertion_clause=None, *args):
            raw_msg = 'Expected topic({{0}}) {assertion_clause}'.format(
                assertion_clause=assertion_clause)
            if len(args) is 2:
                raw_msg += ' {1}'
            return raw_msg

        def exec_assertion(*args):
            raw_msg = _assertion_msg(humanized_method_name, *args)
            if not method(*args):
                raise AssertionError(raw_msg.format(*args))

        def exec_not_assertion(*args):
            raw_msg = _assertion_msg('not {0}'.format(humanized_method_name), *args)
            if method(*args):
                raise AssertionError(raw_msg.format(*args))

        Assertions.registered_assertions[method.__name__] = exec_assertion
        Assertions.registered_assertions['not_{method_name}'.format(method_name=method.__name__)] = exec_not_assertion

        def wrapper(*args, **kw):
            return method(*args, **kw)

        return wrapper


class Expect(object):
    '''This atypical class provides a key part of the preggy testing syntax.

    For example:

        expect(True).to_be_true()

    '''

    def __init__(self, topic):
        self.topic = topic
        self.not_assert = False

    def __getattr__(self, name):
        # common cases
        if name == 'topic':
            return super(Expect, self).__getattr__(name)
        if name == 'Not':
            self.not_assert = not self.not_assert
            return self

        # update `not_` assertions
        if self.not_assert:
            method_name = 'not_{name}'.format(name=name)
        else:
            method_name = name

        # check for unregistered assertions
        if method_name not in Assertions.registered_assertions:
            raise AttributeError('Assertion {method_name} was not found!'.format(method_name=method_name))

        # if program gets this far, then it’s time to perform the assertion. (...FINALLY! ;D)
        def assert_topic(*args, **kw):
            '''Allows instances (topics) to chain calls to `Assertion`s.

            In the following preggy-test snippet:

                expect(topic).to_be_True()

            ...This method is what allows `expect(topic)` to call
            `.to_be_True()` (or some other Assertion).

            '''
            return Assertions.registered_assertions[method_name](self.topic, *args, **kw)

        return assert_topic



# -*- coding: utf-8 -*-
'''preggy error assertions.  For use with `expect()` (see `preggy.core`).
'''

# preggy assertions
# https://github.com/heynemann/preggy

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2013 Bernardo Heynemann heynemann@gmail.com

import inspect

from preggy import assertion, create_assertions


@assertion
def to_be_an_error_like(topic, expected):
    '''Asserts that `topic` is an instance (or subclass) of type
    `expected.
    '''
    if not isinstance(topic, expected):
        raise AssertionError(
            'Expected topic({0}) to be an error of type {1}, but it was a {2}',
            topic, expected, topic.__class__
        )


@assertion
def to_have_an_error_message_of(topic, expected):
    '''Asserts that `topic` has an error message of `expected`.'''
    if str(topic) != expected:
        raise AssertionError('Expected topic({0}) to be an error with message {1}', topic, expected)


@create_assertions
def to_be_an_error(topic):
    '''Asserts that `topic` is an error.'''
    return topic and (isinstance(topic, Exception) or (inspect.isclass(topic) and issubclass(topic, Exception)))



# -*- coding: utf-8 -*-
'''preggy function assertions.  For use with `expect()` (see `preggy.core`).
'''

# preggy assertions
# https://github.com/heynemann/preggy

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2013 Bernardo Heynemann heynemann@gmail.com

import inspect

from preggy import assertion


@assertion
def to_be_a_function(topic):
    '''Asserts that `topic` is a function.'''
    if not (inspect.ismethod(topic) or inspect.isfunction(topic)):
        raise AssertionError('Expected topic({0}) to be a function or a method, but it was a {1}', topic, topic.__class__)


@assertion
def not_to_be_a_function(topic):
    '''Asserts that `topic` is NOT a function.'''
    if inspect.ismethod(topic) or inspect.isfunction(topic):
        raise AssertionError('Expected topic({0}) not to be a function or a method', topic)



# -*- coding: utf-8 -*-
'''preggy 'like' assertions.  For use with `expect()` (see `preggy.core`).
'''


# preggy assertions
# https://github.com/heynemann/preggy

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2013 Bernardo Heynemann heynemann@gmail.com

try:
    from six import string_types, binary_type
except ImportError:
    print("Ignoring six. Probably setup.py installing package.")

import numbers

from preggy import create_assertions


@create_assertions
def to_be_like(topic, expected):
    '''Asserts that `topic` is like (similar to) `expected`. Allows some leeway.'''
    return match_alike(expected, topic)


def match_alike(expected, topic):
    '''Asserts the "like"-ness of `topic` and `expected` according to their types.'''
    if topic is None:
        return expected is None

    if isinstance(topic, string_types + (binary_type, )):
        return compare_strings(expected, topic)

    if isinstance(topic, numbers.Number):
        return compare_numbers(expected, topic)

    if isinstance(topic, (list, tuple)):
        return compare_lists(expected, topic)

    if isinstance(topic, dict):
        return compare_dicts(expected, topic)

    raise RuntimeError('Could not compare {expected} and {topic}'.format(expected=expected, topic=topic))


def compare_strings(expected, topic):
    '''Asserts the "like"-ness of `topic` and `expected` as strings.
    Allows some leeway.  (Strings don't have to exactly match.)

    '''

    if isinstance(topic, (binary_type, )):
        topic = topic.decode('utf-8')

    if isinstance(expected, (binary_type, )):
        expected = expected.decode('utf-8')

    replaced_topic = topic.lower().replace(' ', '').replace('\n', '')
    replaced_expected = expected.lower().replace(' ', '').replace('\n', '')
    return replaced_expected == replaced_topic


def compare_numbers(expected, topic):
    '''Asserts the "like"-ness of `topic` and `expected` as Numbers.'''
    if not isinstance(topic, numbers.Number) or not isinstance(expected, numbers.Number):
        return False
    return float(expected) == float(topic)


def compare_dicts(expected, topic):
    '''Asserts the "like"-ness of `topic` and `expected` as dicts.'''
    return match_dicts(expected, topic) and match_dicts(topic, expected)


def match_dicts(expected, topic):
    '''Asserts the "like"-ness of all keys and values in `topic` and
    `expected`.
    '''
    for k, v in expected.items():
        if not k in topic or not match_alike(topic[k], v):
            return False
    return True


def compare_lists(expected, topic):
    '''Asserts the "like"-ness of `topic` and `expected` as lists.'''
    return match_lists(expected, topic) and match_lists(topic, expected)


def match_lists(expected, topic):
    '''Asserts the "like"-ness each item in of `topic` and `expected`
    (as lists or tuples).

    '''
    for item in expected:
        if isinstance(item, (list, tuple)):
            found = False
            for inner_item in topic:
                if isinstance(inner_item, (list, tuple)) and compare_lists(item, inner_item):
                    found = True
                    break
            if not found:
                return False
        elif not item in topic:
            return False

    return True



# -*- coding: utf-8 -*-

# preggy assertions
# https://github.com/heynemann/preggy

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2013 Bernardo Heynemann heynemann@gmail.com

# flake8: noqa

__version__ = "0.3.9"

from preggy.core import Assertions
assertion = Assertions.assertion
create_assertions = Assertions.create_assertions

from preggy.core import Expect as expect

from preggy.emptiness import *
from preggy.equality import *
from preggy.inclusion import *
from preggy.length import *
from preggy.like import *

from preggy.types.numeric import *
from preggy.types.function import *
from preggy.types.boolean import *
from preggy.types.classes import *
from preggy.types.file import *
from preggy.types.nullable import *
from preggy.types.regexp import *
from preggy.types.errors import *



# -*- coding: utf-8 -*-
'''preggy boolean assertions.  For use with `expect()` (see `preggy.core`).
'''

# preggy assertions
# https://github.com/heynemann/preggy

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2013 Bernardo Heynemann heynemann@gmail.com

from preggy import assertion


@assertion
def to_be_true(topic):
    '''Asserts that the boolean value of `topic` is `True`.'''
    if not bool(topic):
        raise AssertionError("Expected topic({0}) to be truthy", topic)


@assertion
def to_be_false(topic):
    '''Asserts that the boolean value of `topic` is `False`.'''
    if bool(topic):
        raise AssertionError("Expected topic({0}) to be falsy", topic)


@assertion
def not_to_be_true(topic):
    return to_be_false(topic)


@assertion
def not_to_be_false(topic):
    return to_be_true(topic)



# -*- coding: utf-8 -*-
'''preggy length assertions.  For use with `expect()` (see `preggy.core`).
'''


# preggy assertions
# https://github.com/heynemann/preggy

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2013 Bernardo Heynemann heynemann@gmail.com

from preggy import assertion


@assertion
def to_length(topic, expected):
    '''Asserts that `len(topic)` == `expected`.'''
    length = None

    try:
        length = len(topic)
    except AttributeError:
        if hasattr(topic, 'qsize'):
            length = topic.qsize()
    except TypeError:
        if hasattr(topic, 'qsize'):
            length = topic.qsize()

    if length is None:
        raise AssertionError("Could not determine \"{0}\"'s length.".format(topic))

    if length != expected:
        raise AssertionError('Expected "{0}" to have {1} of length, but it has {2}'.format(topic, expected, length))


@assertion
def not_to_length(topic, expected):
    '''Asserts that `len(topic)` != `expected`.'''

    length = None

    try:
        length = len(topic)
    except AttributeError:
        if hasattr(topic, 'qsize'):
            length = topic.qsize()
    except TypeError:
        if hasattr(topic, 'qsize'):
            length = topic.qsize()

    if length is None:
        raise AssertionError("Could not determine \"{0}\"'s length.".format(topic))

    if length == expected:
        raise AssertionError('Expected {0} not to have {1} of length'.format(topic, expected))
